# bash base library version 0.4.13
# Copyright (C)  2015-2019 Heinz Deinhart
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This is a bash library, it has to be sourced not called.
#
# bash-base provides a number of functions and a base setup of settings
# to support clean coding style in bash.
#
# Internal global veriables are prefixed with _BASH_BASE_, internal
# function with _bash_base_

[ "$0" != "$BASH_SOURCE" ] || {
	echo "$0: this script must be sourced, not called" >&2
	exit 1
}

#@@ error messages

#@ error-codes (bash-base)
#   1 .. unkown internal error
#   2 .. illegal number of arguments in function call
#   3 .. illegal variable name
#   4 .. illegal function arguments
#   5 .. error sourcing file
#   6 .. no such function
#   7 .. variable specified as argument undefined
#   8 .. no such file
#   9 .. no access rights
#  9x .. errors local to each function

#@ bb_error_string <ret-vn> <error-code>
# errors: 2, 3
function bb_error_string() {
    (( $# == 2 )) || {
	bb_ierr "bb_error_string: called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _err_code="$2"

    bb_validate_vn bb_error_string "$_ret_vn" || return 3

    local _msg
    case "$_err_code" in
	0) _msg="no error";;
	1) _msg="unkown internal error";;
	2) _msg="illegal number of arguments in function call";;
	3) _msg="illegal variable name";;
	4) _msg="illegal function arguments";;
	5) _msg="error sourcing file";;
	6) _msg="no such function";;
	7) _msg="variable specified as argument undefined";;
	8) _msg="no such file";;
	9) _msg="no access rights";;
	*) _msg="bb_error_string: unknown error code $_err_code";;
    esac
    printf -v "$_ret_vn" "%s" "$_msg"
}

#@@ generic tools

#@ bb_validate_vn <caller-name> <vn> [..]
# validates one or more variable names
# valid variable names match the following regex: ^[[:alpha:]][[:alnum:]_]*$
# if caller-name is non empty a log msg with level warning will be emitted
# errors: 2, 3
function bb_validate_vn() {
    (( $# > 1 )) || {
	bb_ierr "bb_validate_vn: called with $# instead of 2 or more args"
	return 2
    }

    local _caller="$1"
    shift

    while (( $# > 0 ))
    do
	[[ "$1" =~ ^[[:alpha:]][[:alnum:]_]*$ ]] || {
	    [ -z "$_caller" ] || bb_msg warning \
		"bb_validate_vn($_caller): illegal variable name: $1"
	    return 3
	}
	shift
    done
}

#@ bb_validate_regex <caller-name> <regex> <arg> [..]
# validates one or more arguments against specified regex
# if caller-name is non empty a log msg with level warning will be emitted
# errors: 2, 3
function bb_validate_regex() {
    (( $# > 2 )) || {
	bb_ierr "bb_validate_regex: called with $# instead of 3 or more args"
	return 2
    }

    local _caller="$1"
    local _regex="$2"
    shift 2

    while (( $# > 0 ))
    do
	[[ "$1" =~ $_regex ]] || {
	    [ -z "$_caller" ] || bb_msg warning \
		"bb_validate_regex($_caller): illegal argument: $1"
	    return 3
	}
	shift
    done
}

#@ bb_validate_uint <caller-name> <uint> [..]
# validates if one or more arguments are unsigned integers
# if caller-name is non empty a log msg with level warning will be emitted
# errors: 2, 3
function bb_validate_uint() {
    (( $# > 1 )) || {
	bb_ierr "bb_validate_uint: called with $# instead of 2 or more args"
	return 2
    }

    local _caller="$1"
    shift

    while (( $# > 0 ))
    do
	[[ "$1" =~ ^[[:digit:]]+$ ]] || {
	    [ -z "$_caller" ] || bb_msg warning \
		"bb_validate_uint($_caller): illegal unsigned integer: $1"
	    return 3
	}
	shift
    done
}

#@ bb_validate_int <caller-name> <int> [..]
# validates if one or more arguments are integers
# if caller-name is non empty a log msg with level warning will be emitted
# errors: 2, 3
function bb_validate_int() {
    (( $# > 1 )) || {
	bb_ierr "bb_validate_int: called with $# instead of 2 or more args"
	return 2
    }

    local _caller="$1"
    shift

    while (( $# > 0 ))
    do
	[[ "$1" =~ ^[[:digit:]-][[:digit:]]*$ ]] || {
	    [ -z "$_caller" ] || bb_msg warning \
		"bb_validate_int($_caller): illegal integer: $1"
	    return 3
	}
	shift
    done
}

# bb_executables_in_path <caller-name> <function-name> [..]
# checks if specified executables present (absolute or relative) or if they
# are in path or they are defined functions
# errors: 2, 8
function bb_executables_in_path() {
    (( $# > 1 )) || {
	bb_ierr "bb_funs_exist: called with $# instead of 2 or more args"
	return 2
    }

    local _caller="$1"
    shift

    while (( $# > 0 ))
    do
	local _ex="$1"
	shift

	if declare -f "$_ex" >/dev/null
	then
	    continue # function
	fi

	if [[ "$_ex" =~ ^((\.?\./)|/) ]]
	then
	    # absolute or relative path included
	    if [[ -x "$_ex" ]]
	    then
		continue
	    fi
	else
	    local _found_ex
	    _found_ex="$(which "$_ex" 2>/dev/null || true)"

	    if [[ "$_found_ex" ]]
	    then
		continue
	    fi
	fi

	bb_msg debug "bb_funs_exist($_caller): executable not found: $_ex"
	return 8
    done
}

#@@ string tools

#@ bb_string_compare <string1> <string2>
# compare if two strings are equal, return 0 if they are
# errors: 2, 91 (not equal)
function bb_string_compare() {
   (( $# == 2 )) || {
	bb_ierr "bb_string_compare called with $# instead of 2 args"
	return 2
    }

   [[ "$1" == "$2" ]] || return 91
}

#@@ array tools

#@ bb_array_contains_value <array-vn> <string>
# check if array contains value <string>, return 0 if found
# errors: 2, 91 (not found)
function bb_array_contains_value() {
    (( $# == 2 )) || {
	bb_ierr "bb_array_contains called with $# instead of 2 args"
	return 2
    }

    local _array_vn="$1"
    local _str="$2"
    local -a bb_array_contains_value__fr

    bb_array_filter bb_array_contains_value__fr "$_array_vn" \
	bb_string_compare "$_str"

    (( ${#bb_array_contains_value__fr[*]} > 0 )) || return 91
}

#@ bb_array_compare <src1-vn> <src2-vn>
# deep compare two arrays, return 0 if identically
# errors: 2, 3, 91 (not identically)
function bb_array_compare() {
    (( $# == 2 )) || {
	bb_ierr "bb_array_compare called with $# instead of 2 args"
	return 2
    }

    local _src1_vn="$1"
    local _src2_vn="$2"

    bb_validate_vn bb_array_compare "$_src1_vn" "$_src2_vn" || return 3

    local _src1 _src2
    _src1="$(declare -p "$_src1_vn")"
    _src1="${_src1#*=}"
    _src2="$(declare -p "$_src2_vn")"
    _src2="${_src2#*=}"

    [[ "$_src1" == "$_src2" ]] || return 91
}

#@ bb_array_copy_without_vals <dst-vn> <src-vn> <val> [..]
# copies array named srv-vn into array called dst-vn while skipping all
# specified vals
# errors: 2
function bb_array_copy_without_vals() {
    (( $# > 2 )) || {
	bb_ierr "bb_array_copy_without_vals called with $#" \
	    "instead of 3 or more args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"
    shift 2

    bb_validate_vn bb_array_copy_without_vals "$_dst_vn" "$_src_vn"

    local -a _src_keys
    eval "_src_keys=(\"\${!$_src_vn[@]}\")"

    # clear destination array
    eval "$_dst_vn=()"

    # empty without anything is empty
    (( ${#_src_keys[@]} > 0 )) || return

    local _key _val _ref _element
    for _key in "${_src_keys[@]}"
    do
	_ref="${_src_vn}[$_key]"
	_val="${!_ref}"
	for _element in "$@"
	do
	    [[ "$_val" != "$_element" ]] || continue 2
	done
	eval "${_dst_vn}[$_key]=\"\${_val}\""
    done
}

#@ bb_array_remove_vals <array-vn> <val> [..]
# removes specified vals from array named array-vn
# errors: 2
function bb_array_remove_vals() {
    (( $# > 1 )) || {
	bb_ierr "bb_array_remove_vals called with $# instead of 1 or more args"
	return 2
    }

    local _array_vn="$1"
    shift

    bb_validate_vn bb_array_remove_vals "$_array_vn"

    # shellcheck disable=SC2034
    local -a bb_array_remove_vals__tmp_a
    bb_array_copy_without_vals bb_array_remove_vals__tmp_a "$_array_vn" "$@"
    bb_array_copy "$_array_vn" bb_array_remove_vals__tmp_a
}

#@ bb_array_reverse <ret-vn> <src-vn>
# returns reversed array
# same indices are used, only data is swapped (0 with n, 1 with (n-1), ..)
# erors: 2, 3
function bb_array_reverse() {
    (( $# == 2 )) || {
	bb_ierr "bb_array_reverse called with $# instead of 2 args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"

    bb_validate_vn bb_array_reverse "$_dst_vn" "$_src_vn" || return 3

    local -a _keys
    eval "_keys=(\"\${!$_src_vn[@]}\")"

    # clear destination array
    eval "$_dst_vn=()"

    local _i _j _ks _kt
    for (( _j = 0, _i = ${#_keys[@]} - 1; _i >= 0; _i--, _j++ ))
    do
	_ks="${_keys[$_i]}"
	_kt="${_keys[$_j]}"

	eval "${_dst_vn}[$_kt]=\"\${$_src_vn[$_ks]}\""
    done
}

#@ bb_array_iter <array-vn> <iterfun-name> [<xarg> [..]]
# iter-fun must take: <ret-vn> <array-value> [<xarg> [..]]
# if target array is not empty contents will be merged in
# errors: 2, 3, 8
function bb_array_iter() {
    (( $# > 1 )) || {
	bb_ierr "bb_array_iter called with $# instead of 2 or more args"
	return 2
    }

    local _src_vn="$1"
    local _iter_fun="$2"
    shift 2

    bb_validate_vn bb_array_iter "$_src_vn" || return 3
    bb_executables_in_path bb_array_iter "$_iter_fun" || return 8

    local -a _keys
    local _key _ref
    eval "_keys=(\"\${!$_src_vn[@]}\")"

    # no need to process empty arrays
    (( ${#_keys[@]} > 0 )) || return 0

    for _key in "${_keys[@]}"
    do
	_ref="${_src_vn}[$_key]"
	"$_iter_fun" "${!_ref}" "$@"
    done
}

#@ bb_array_map <ret-vn> <src-vn> <mapfun-name> [<xarg> [..]]
# map-fun must take: <ret-vn> <src-value> [<xarg> [..]]
# if target array is not empty contents will be merged in
# errors: 2, 3, 8
function bb_array_map() {
    (( $# > 2 )) || {
	bb_ierr "bb_array_map called with $# instead of 3 or more args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"
    local _map_fun="$3"
    shift 3

    bb_validate_vn bb_array_map "$_src_vn" "$_dst_vn" || return 3
    bb_executables_in_path bb_array_map "$_map_fun" || return 8

    local -a _keys
    local bb_array_map__cache _key _ref
    eval "_keys=(\"\${!$_src_vn[@]}\")"

    # clear destination array
    eval "$_dst_vn=()"

    # no need to process empty arrays
    (( ${#_keys[@]} > 0 )) || return 0

    for _key in "${_keys[@]}"
    do
	_ref="${_src_vn}[$_key]"
	"$_map_fun" bb_array_map__cache "${!_ref}" "$@"
	eval "${_dst_vn}[$_key]=\"\${bb_array_map__cache}\""
    done
}

#@ bb_array_map2 <ret-vn> <src1-vn> <src2-vn> <mapfun-name> [<xarg> [..]]
# map-fun must take: <ret-vn> <src1-value> <src2-value> [<xarg> [..]]
# if target array is not empty contents will be merged in
# note: this function processes elements one by one and ignores keys
# errors: 2, 3, 8, 91 (src1 and src2 have different length)
function bb_array_map2() {
    (( $# > 3 )) || {
	bb_ierr "bb_array_map called with $# instead of 3 or more args"
	return 2
    }

    local _dst_vn="$1"
    local _src1_vn="$2"
    local _src2_vn="$3"
    local _map_fun="$4"
    shift 3

    bb_validate_vn bb_array_map2 "$_src1_vn" "$_src2_vn" "$_dst_vn" || \
	return 3
    bb_executables_in_path bb_array_map2 "$_map_fun" || return 8

    local -a _keys1
    local -a _keys2
    eval "_keys1=(\"\${!$_src1_vn[@]}\")"
    eval "_keys2=(\"\${!$_src2_vn[@]}\")"

    (( ${#_keys1[@]} == ${#_keys2[@]} )) || return 91

    # clear destination array
    eval "$_dst_vn=()"

    # no need to process empty arrays
    (( ${#_keys1[@]} > 0 )) || return 0

    local _index _ref1 _ref2 _key1 _key2
    # shellcheck disable=SC2034
    local bb_array_map__cache
    for (( _index=0; _index < ${#_keys1[@]}; _index++ ))
    do
	_key1="${_keys1[$_index]}"
	_key2="${_keys2[$_index]}"
	_ref1="${_src1_vn}[$_key1]"
	_ref2="${_src2_vn}[$_key2]"
	"$_map_fun" bb_array_map__cache "${!_ref1}" "${!_ref2}" "$@"
	eval "${_dst_vn}[$_index]=\"\${bb_array_map__cache}\""
    done
}

#@ bb_array_fold_right <ret-vn> <fold-fun> <src-vn> <start-val> [<xarg> [..]]
# folds an array from left to right
# fold-fun must take: <ret-vn> <element> <old-result> [<xarg> [..]]
# errors: 2, 3, 8
function bb_array_fold_right() {
    (( $# > 3 )) || {
	bb_ierr "bb_array_fold_right called with $# instead of 4 or more args"
	return 2
    }

    local _ret_vn="$1"
    local _fold_fun="$2"
    local _src_vn="$3"
    local _start="$4"
    shift 4

    bb_validate_vn bb_array_map "$_src_vn" "$_ret_vn" || return 3
    bb_executables_in_path bb_array_fold_right "$_fold_fun" || return 8

    # shellcheck disable=SC2034
    local -a bb_array_fold_right__keys
    eval "bb_array_fold_right__keys=(\"\${!$_src_vn[@]}\")"

    # fold right wants to start from back
    local -a bb_array_fold_right__rkeys
    bb_array_reverse bb_array_fold_right__rkeys bb_array_fold_right__keys

    local bb_array_fold_right__cache _key _ref
    bb_array_fold_right__cache="$_start"
    for _key in "${bb_array_fold_right__rkeys[@]}"
    do
	_ref="${_src_vn}[$_key]"
	"$_fold_fun" bb_array_fold_right__cache \
	    "${!_ref}" "$bb_array_fold_right__cache" "$@"
    done

    printf -v "$_ret_vn" "%s" "$bb_array_fold_right__cache"
}

#@ bb_array_fold_left <ret-vn> <fold-fun> <start-val> <src-vn> [<arg> [..]]
# folds an array from left to right
# fold-fun must take: <ret-vn> <old-result> <element> [<arg> [..]]
# errors: 2, 3, 8
function bb_array_fold_left() {
    (( $# > 3 )) || {
	bb_ierr "bb_array_fold_left called with $# instead of 4 or more args"
	return 2
    }

    local _ret_vn="$1"
    local _fold_fun="$2"
    local _start="$3"
    local _src_vn="$4"
    shift 4

    bb_validate_vn bb_array_map "$_src_vn" "$_ret_vn" || return 3
    bb_executables_in_path bb_array_fold_left "$_fold_fun" || return 8

    local -a _keys
    local bb_array_fold_left__cache _key _ref
    eval "_keys=(\"\${!$_src_vn[@]}\")"

    bb_array_fold_left__cache="$_start"
    for _key in "${_keys[@]}"
    do
	_ref="${_src_vn}[$_key]"
	"$_fold_fun" bb_array_fold_left__cache \
	    "$bb_array_fold_left__cache" "${!_ref}" "$@"
    done

    printf -v "$_ret_vn" "%s" "$bb_array_fold_left__cache"
}

#@ bb_array_filter <ret-vn> <src-vn> <filterfun-name> [<xarg> [..]]
# filterfun must take: <element> [<xarg> [..]]
# and shall return 0 if the element is to be kept in output array
# errors: 2, 3, 8
function bb_array_filter() {
    (( $# > 2 )) || {
	bb_ierr "bb_array_filter called with $# instead of 3 or more args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"
    local _filter_fun="$3"
    shift 3

    bb_validate_vn bb_array_filter "$_src_vn" "$_dst_vn" || return 3
    bb_executables_in_path bb_array_filter "$_filter_fun" || return 8

    local -a _keys
    local _key _ref

    eval "_keys=(\"\${!$_src_vn[@]}\")"

    # clear destination array
    eval "$_dst_vn=()"

    # no need to process empty arrays
    (( ${#_keys[@]} > 0 )) || return 0

    for _key in "${_keys[@]}"
    do
	_ref="${_src_vn}[$_key]"
	if "$_filter_fun" "${!_ref}" "$@"
	then
	    eval "${_dst_vn}[$_key]=\"\${!_ref}\""
	fi
    done
}


#@ bb_array_copy <dst-vn> <src-vn>
# copys an array, target array must exist and if non empty will be merged
# errors: same as bb_array_map
function bb_array_copy() {
    (( $# == 2 )) || {
	bb_ierr "bb_array_copy called with $# instead of 2 args"
	return 2
    }

    local _dst_vn="$1"
    local _src_vn="$2"

    bb_validate_vn bb_array_copy "$_src_vn" "$_dst_vn" || return 3

    bb_array_map "$_dst_vn" "$_src_vn" bb_mapfun_identity
}

#@ bb_array_length <ret-vn> <array-vn>
# returns the length of an array
# errors: 2, 3
function bb_array_length() {
    local _ret_vn="$1"
    local _src_vn="$2"

    (( $# == 2 )) || {
	bb_ierr "bb_array_length called with $# instead of 2 args"
	return 2
    }

    bb_validate_vn bb_array_length "$_ret_vn" "$_src_vn" || return 3

    local -a _keys
    eval "_keys=(\"\${!$_src_vn[@]}\")"

    printf -v "$_ret_vn" "%i" "${#_keys[@]}"
}

#@ bb_mapfun_identity <dst-vn> <input-value>
# the identity fun
# errors: 2, 3
function bb_mapfun_identity() {
    (( $# == 2 )) || {
	bb_ierr "bb_mapfun_identity called with $# instead of 2 args"
	return 2
    }

    bb_validate_vn bb_mapfun_identity "$1" || return 3

    printf -v "$1" "%s" "${2}"
}

#@ bb_array_join_by <ret-vn> <join-string> [<element0> [..]]
# joins array element with join-string, returns "" if no elements specified
# errors: 2, 3
function bb_array_join_by() {
    (( $# > 1 )) || {
	bb_ierr "bb_array_join_by called with $# instead of 2 or more args"
	return 2
    }

    local _ret_vn="$1"
    local _join_str="$2"
    shift 2

    bb_validate_vn bb_array_join_by "$_ret_vn" || return 3

    # return "" if we have nothing to join
    (( $# > 0 )) || {
	printf -v "$_ret_vn" "%s" ""
	return
    }

    local _result="$1"
    shift

    while (( $# > 0 ))
    do
	_result="${_result}${_join_str}$1"
	shift
    done

    printf -v "$_ret_vn" "%s" "$_result"
}

#@ bb_array_of_string <dst-vn> <string> [ <batch-size> ]
# converts a string into an array of strings with size batch-size and copy
# it into the array called dst-vn
# the default batch-size is 1
# note: the last batch may be shorter than batch-size
function bb_array_of_string() {
    (( $# == 2 || $# == 3 )) || return 123

    local _dst_vn="$1"
    local _input_str="$2"
    local _batch_size="${3-1}"

    # clear destination array
    eval "$_dst_vn=()"

    # nothing to do when input string is empty
    (( ${#_input_str} > 0 )) || return 0

    local i c=0
    for (( i=0; i < ${#_input_str}; i = i + _batch_size, c++ ))
    do
	# shellcheck disable=SC2016
	eval "$_dst_vn[$c]="'${_input_str:$i:$_batch_size}'
    done
}

#@@ hexadecimal

#@ bb_hex_validate <caller-name> <hex-string> [..]
# validates one or more hex strings
# valid variable names match the following regex: ^[[:xdigit:]]*$
# errors: 2, 91 (illegal hex code)
function bb_hex_validate() {
    (( $# > 1 )) || {
	bb_ierr "bb_hex_validate: called with $# instead of 2 or more args"
	return 2
    }

    local _caller="$1"
    shift

    while (( $# > 0 ))
    do
	[[ "$1" =~ ^[[:xdigit:]]*$ ]] || {
	    bb_msg warning \
		"bb_hex_validate($_caller): illegal hex code: $1"
	    return 91
	}
	shift
    done
}

#@ bb_hex_read <ret-vn>
# converts binary data read from stdin to its hex representation
# make sure it does not run in subshell, call it e.g.: bb_hex_read x < <(date)
# requires: /usr/bin/od
# errors: 2, 3, 91 (od not found)
function bb_hex_read() {
    (( $# == 1 )) || {
	bb_ierr "bb_hex_read called with $# args instead of 1"
	return 2
    }

    local _ret_vn="$1"

    bb_validate_vn bb_hex_read "$_ret_vn" || return 3

    local _od
    if [ -x /usr/bin/od ]
    then
	_od=/usr/bin/od
    else
	return 91
    fi

    local _ret_str=
    while read line
    do
	printf -v _ret_str "%s%s" "$_ret_str" "${line//[[:space:]]/}"
    done < <("$_od" -A n -v -t x1)

   printf -v "$_ret_vn" "%s" "$_ret_str"
}

#@ bb_bes_of_hex <ret-vn> <hex>
# converts hex to backslash escape sequence, which can be converted to data
# by passing to: printf "%b"
# hex is interpreted as stream of bytes (means it must be of even length)
# errors: 2, 91 (hex is of odd length)
function bb_hex_to_bes() {
    (( $# == 2 )) || {
	bb_ierr "bb_hex_to_bes called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _input="$2"

    bb_validate_vn bb_hex_to_bes "$_ret_vn"

    (( ${#_input} > 0 )) || { # no input leads to no output
	eval "$_ret_vn="
	return
    }

    (( ${#_input} % 2 == 0 )) || return 91

    local -a bb_hex_to_bes__inar bb_hex_to_bes__joined
    bb_array_of_string bb_hex_to_bes__inar "$_input" 2
    bb_array_join_by bb_hex_to_bes__joined "\\x" "${bb_hex_to_bes__inar[@]}"
    printf -v "$_ret_vn" "%s" "\\x$bb_hex_to_bes__joined"
}

#@ bb_hex_write <hex>
# writes hex encoded data as binary data to stdout
# hex is interpreted as stream of bytes (means it must be of even length)
# errors: 2, 91 (hex is of odd length), 92 (invalid hex char)
function bb_hex_write() {
    (( $# == 1 )) || {
	bb_ierr "bb_hex_write called with $# instead of 1 arg"
	return 2
    }

    local _hex="$1"
    bb_hex_validate bb_hex_write "$_hex" || return 92

    local bb_hex_write__bes

    bb_hex_to_bes bb_hex_write__bes "$_hex"
    printf "%b" "$bb_hex_write__bes"
}

#@ bb_hex_xor <ret-vn> <hex1> <hex2>
# returns hex1 xor hex2 as hex
# errors: 2, 3, 91 (input length vary), 92 (invaid hex char)
function bb_hex_xor() {
    (( $# == 3 )) || {
	bb_ierr "bb_hex_xor called with $# instead of 3 args"
	return 2
    }

    local _ret_vn="$1"
    local _hex1="$2"
    local _hex2="$3"
    bb_hex_validate bb_hex_xor "$_hex1" "$_hex2" || return 92

    local _len
    _len=${#_hex1}
    (( _len == ${#_hex2} )) || return 91

    local _i _ret_str=
    for (( _i=0; _i < _len; _i++ ))
    do
	printf -v _ret_str "%s%x" \
	    "$_ret_str" $(( 0x${_hex1:$_i:1} ^ 0x${_hex2:$_i:1} ))
    done

    printf -v "$_ret_vn" "%s" "$_ret_str"
}

#@@ error handling: handle internal errors in a configurable way

#@ bb_ierr <msg> [..]
# call this on internal error, depending on mode it does:
#  mode "fatal": pass msg to bb_fatal
#  other mode: print error msg with specified loglevel
function bb_ierr() {
    local _msg

    if (( $# < 1 ))
    then
	_msg="bb_ierr called w/o message"
    else
	_msg="$*"
    fi

    if [[ "$bb__ierrmode" == fatal ]]
    then
	bb_fatal "$_msg"
    else
	bb_msg "$bb__ierrmode" "$_msg"
    fi
}

#@ bb_ierr_set_mode [fatal|<msg-loglevel>]
# default is "err"
function bb_ierr_set_mode() {
    local _m="${1-err}"

    if [[ "$_m" =~ ^(fatal|debug|info|notice|warning|err|crit|alert|emerg)$ ]]
    then
	bb__ierrmode="$_m"
    else
	bb_ierr "bb_ierr_set_mode called with illegal argument: $_m"
    fi
}

#@ bb_fatal <message> [..]
# end program execution after logging message prepended with 'fatal error'
# since it produces some lines of output it is meant for internal errors
# only, if you want to abort your script with a message use bb_quit
function bb_fatal() {
    bb_msg emerg "fatal error: $*"
    bb_quit 1
}

#@@ msg: log/show messages

#@ bb_msg_levelname_of_string <ret-vn> <string>
# returns level as string if input was integer
# else returns string if it is a valid level
# errors: 2, 3, 4
function bb_msg_levelname_of_string() {
    (( $# == 2 )) || {
	bb_ierr "bb_msg_levelname_of_string called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _str="$2"

    bb_validate_vn bb_msg_levelname_of_string "$_ret_vn" || return 3

    if [[ "$_str" =~ ^[01234567]$ ]]
    then
	local -a _lookup=(emerg alert crit err warning notice info debug)

	printf -v "$_ret_vn" "%s" "${_lookup[$_str]}"
    else
	if [[ "$_str" =~ ^(debug|info|notice|warning|err|crit|alert|emerg)$ ]]
	then
	    printf -v "$_ret_vn" "%s" "$_str"
	else
	    # invalid argument
	    return 4
	fi
    fi
}

#@ bb_msg_int_of_level
# convert level name to int or return int unmodified if in valid range
# errors: 2, 3, 4
function bb_msg_int_of_level() {
    (( $# == 2 )) || {
	bb_ierr "bb_msg_int_of_level called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _level="$2"

    bb_validate_vn bb_msg_int_of_level "$_ret_vn" || return 3

    local _result=7 # debug is default

    if [[ "$2" =~ ^[01234567]$ ]]
    then
	_result="$_level"
    else
	case "$_level" in
	    emerg) _result=0;;
	    alert) _result=1;;
	    crit) _result=2;;
	    err) _result=3;;
	    warning) _result=4;;
	    notice) _result=5;;
	    info) _result=6;;
	    debug) _result=7;;
	    *) return 4
	esac
    fi

    printf -v "$_ret_vn" "%i" "$_result" # we are already int
}

#@ bb_msg_set_target <target-name> <type>[:<arg>] <min-level>[:<max-level]
#                   <prefix> [[no]trace]
# where type can be one of: stdout, stderr, file, syslog
# and arg is filename for file, facility for syslog
# level is one of: debug, info, notice, warning, err, crit, alert, emerg
# prefix can be: none, timestamp, prog, progpid, full
# if notrace is specified, no stack traces will be print to that target
# notrace is the default
# note: target-name must match regex ^[[:alpha:]][[:alnum:]_]*$
# errors: 2, 4
function bb_msg_set_target() {
    (( $# == 4 || $# == 5 )) || {
	bb_ierr "bb_msg_set_target called with $# instead of 5 or 6 args"
	return 2
    }

    local _tn="$1"

    [[ "$_tn" =~ ^[[:alpha:]][[:alnum:]_]*$ ]] || {
	bb_msg err "bb_msg_set_target: illegal target-name: $_tn"
	return 4
    }

    local _type="$2"
    local _levelstr="$3"
    local _prefix="$4"
    local _prtr="${5-notrace}"

    local bb_msg_set_target__minlevel bb_msg_set_target__maxlevel

    if [[ "$_levelstr" =~ : ]]
    then
	bb_msg_int_of_level bb_msg_set_target__minlevel "${_levelstr%%:*}" || {
	    bb_ierr "bb_msg_set_target called with illegal min-level:" \
		"${_levelstr%%:*}"
	    return 4
	}
	bb_msg_int_of_level bb_msg_set_target__maxlevel "${_levelstr#*:}" || {
	    bb_ierr "bb_msg_set_target called with illegal max-level:" \
		"${_levelstr#*:}"
	    return 4
	}
    else
	bb_msg_int_of_level bb_msg_set_target__minlevel "${_levelstr}" || {
	    bb_ierr "bb_msg_set_target called with illegal min-level:" \
		"${_levelstr}"
	    return 4
	}
	bb_msg_set_target__maxlevel=0
    fi

    # extract arg from type if present
    local _arg=
    if [[ "$_type" =~ : ]]
    then
	_arg="${_type#*:}"
	_type="${_type%%:*}"
    fi

    case "$_type" in
	stdout|stderr)
	    [ -z "$_arg" ] || {
		bb_ierr "bb_msg_set_target type must be empty for stdin/out"
		return 4
	    }
	    ;;
	file) ;;
	syslog)
	    if [ -z "$_arg" ]
	    then
		_arg=syslog # default to syslog
	    else
		[[ "$_arg" =~ ^(auth(priv)?|cron|daemon|ftp|lpr|mail|news|syslog|user|uucp|local[01234567]|)$ ]] || {
		    bb_ierr "bb_msg_set_target illegal syslog arg: $_arg"
		    return 4
		}
	    fi
	    ;;
	*)
	    bb_ierr "bb_msg_set_target called with illegal type: $_type"
	    return 4
    esac

    [[ "$_prefix" =~ ^(none|timestamp|prog|full)$ ]] || {
	bb_ierr "bb_msg_set_target called with illegal prefix"
	return 4
    }

    case "$_prtr" in
	trace) _prtr=yes;;
	notrace) _prtr=no;;
	*)
	    bb_ierr "bb_msg_set_target called with illegal arg 6: $_prtr"
	    return 4
    esac

    bb_msgtargets__type["$_tn"]="$_type"
    bb_msgtargets__arg["$_tn"]="$_arg"
    bb_msgtargets__min_level["$_tn"]="$bb_msg_set_target__minlevel"
    bb_msgtargets__max_level["$_tn"]="$bb_msg_set_target__maxlevel"
    bb_msgtargets__prefix["$_tn"]="$_prefix"
    bb_msgtargets__print_traces["$_tn"]="$_prtr"
}

#@ bb_msg_del_target <target-name>
# errors: 2
function bb_msg_del_target() {
    (( $# == 1 )) || {
	bb_ierr "bb_msg_del_target called with $# instead of 1 arg"
	return 2
    }

    unset bb_msgtargets__type["$1"]
    unset bb_msgtargets__arg["$1"]
    unset bb_msgtargets__min_level["$1"]
    unset bb_msgtargets__max_level["$1"]
    unset bb_msgtargets__prefix["$1"]
    unset bb_msgtargets__print_traces["$1"]
}

#@ bb_msg_clear_targets
# remove all message targets
# errors: 2
# shellcheck disable=SC2120
function bb_msg_clear_targets() {
    (( $# == 0 )) || {
	bb_ierr "bb_msg_del_target called with $# instead of no arg"
	return 2
    }

    unset bb_msgtargets__type
    unset bb_msgtargets__arg
    unset bb_msgtargets__min_level
    unset bb_msgtargets__max_level
    unset bb_msgtargets__prefix
    unset bb_msgtargets__print_traces

    declare -g -A bb_msgtargets__type=()
    declare -g -A bb_msgtargets__arg=()
    declare -g -A bb_msgtargets__min_level=()
    declare -g -A bb_msgtargets__max_level=()
    declare -g -A bb_msgtargets__prefix=()
    declare -g -A bb_msgtargets__print_traces=()
}

# msgr helper functions take: <msg> <msg-prefix> <msg-level-string> <type-arg>
function _bb_msg_to_syslog() {
    logger -p "$4.$3l" "${2-}$1"
}
function _bb_msg_to_stdout() {
    echo "${2-}$1"
}
function _bb_msg_to_stderr() {
    echo "${2-}$1" >&2
}
function _bb_msg_to_file() {
    echo "${2-}$1" >> "$4"
}

#@ bb_msg <level> <string> [..]
# logs messages to configured targets
# note: will return 0, even if level is illegal
# errors: 2, 4
function bb_msg() {
    (( $# > 1 )) || {
	bb_ierr "bb_msg called with $# instead of 2 or more args"
	return 2
    }

    local bb_msg__level bb_msg__levelname
    if bb_msg_int_of_level bb_msg__level "$1"
    then
	_level="$bb_msg__level"
    else
	bb_ierr "bb_msg called with illegal message level: $1"
	return 4
    fi
    bb_msg_levelname_of_string bb_msg__levelname "$1" || {
	bb_ierr "bb_msg called with illegal message level: $1"
	return 4
    }
    shift

    local _tname _targ _message_prefix
    local _timestamp _prog _progpid
    _timestamp="$(date '+%F %T')"
    _prog="${0##*/}"
    _progpid="$_prog[${BASHPID}]"

    for _tname in "${!bb_msgtargets__type[@]}"
    do
	(( "${bb_msgtargets__min_level["$_tname"]}" >= bb_msg__level )) || \
	    continue
	(( "${bb_msgtargets__max_level["$_tname"]}" <= bb_msg__level )) || \
	    continue

	case "${bb_msgtargets__prefix[$_tname]}" in
	    none) _message_prefix=;;
	    timestamp) _message_prefix="$_timestamp: ";;
	    prog) _message_prefix="$_prog: ";;
	    progpid) _message_prefix="$_progpid: ";;
	    full) _message_prefix="$_timestamp $_progpid: ";;
	esac

	_targ="${bb_msgtargets__arg[$_tname]}"

	local _type="${bb_msgtargets__type[$_tname]}"
	local _outfun
	if [[ "$_type" =~ ^syslog|std(out|err)|file$ ]]
	then
	    _outfun=_"bb_msg_to_$_type"
	    else
	    bb_ierr warning "bb_msg encountered illegal type: $_type"
	fi
	"$_outfun" "$*" "$_message_prefix" "$bb_msg__levelname" "$_targ" ||  {
	    echo "${$0##*/}: bb_msg: failed to write message to file" \
		"($_targ): $_message_prefix$*" >&2
	}

	if [[ "${bb_msgtargets__print_traces[$_tname]}" == "yes" ]]
	then
	    # shellcheck disable=SC2034
	    local bb_msg__strace
	    bb_stacktrace_create bb_msg__strace 2
	    bb_array_iter bb_msg__strace "$_outfun" \
		"$_message_prefix" "$bb_msg__levelname" "$_targ"
	fi
    done
}

#@ bb_msgf <level> <format> [<format-arg> [..]]
# logs printf style messages to configured targets
# note: will return 0, even if level is illegal
# errors: 2, 4
function bb_msgf() {
    (( $# > 1 )) || {
	bb_ierr "bb_msgf called with $# instead of 2 or more args"
	return 2
    }

    local _level="$1"
    local _format="$2"
    shift 2

    local _message
    # shellcheck disable=SC2059
    printf -v _message "$_format" "$@" || {
	bb_msg warning "bb_msgf illegal format-string($_format) or args: $*"
	return 4
    }
    bb_msg "$_level" "$_message"
}

#@ bb_msg_stacktrace <ret-vn> [<skip>]
# create a human readable stacktrace, excluding this function and
# returns it as array of strings
# optional arg skip can be used to manually configure how many inner
# levels will not be displayed
# errors: 2, 3
function bb_stacktrace_create() {
    (( $# == 1 || $# == 2 )) || {
	bb_ierr "bb_stacktrace_create called with $# instead of 1 or 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _skip="${2-1}"

    bb_validate_vn bb_stacktrace_create "$_ret_vn" || return 3

    local -a bb_stacktrace_create__strace=()
    # shellcheck disable=SC2034
    local _i _str _intro="Execution at"

    for (( _i=_skip; _i < (${#FUNCNAME[@]} - 1); _i++ ))
    do
	printf -v _str "%s file \"%s\", line %i, function \"%s\"" "$_intro" \
	    "${BASH_SOURCE[$_i]}" "${BASH_LINENO[$_i-1]}" "${FUNCNAME[$_i]}"
	# shellcheck disable=SC2034
	bb_stacktrace_create__strace[$((_i-1))]="$_str"
	_intro="Called from"
    done
    _i=$(( ${#FUNCNAME[@]} - 1 ))
    printf -v _str "%s file \"%s\", line %i" "Called from" \
	"${BASH_SOURCE[$_i]}" "${BASH_LINENO[$_i-1]}"
    # shellcheck disable=SC2034
    bb_stacktrace_create__strace[$_i]="$_str"

    bb_array_copy "$_ret_vn" bb_stacktrace_create__strace
}


#@@ log functions (deprecated, use msg instead)

#@ bb_logf <message> [..]
# logs message file name in $LOG_FILE if non zero
function bb_logf() {
    local _now

    _now="$(date '+%F %T')"

    if [ "${LOG_FILE+isset}" == "isset" ]
    then
	if [ -z "$LOG_FILE" ]
	then
	    return
	fi
	echo "$_now ${0##*/}[${BASHPID}]: $* <<BB WARNING: bb_logf deprecated>>" >> "$LOG_FILE" || true
    fi
}

#@ bb_log <message> [..]
# logs message to stderr and if define to file name in $LOG_FILE
function bb_log() {
    local _now

    _now="$(date '+%F %T')"
    echo "$_now ${0##*/}[${BASHPID}]: $* <<BB WARNING: bb_log deprecated>>" \
	>&2 || true
    if [ "${LOG_FILE+isset}" == "isset" ]
    then
	if [ -z "$LOG_FILE" ]
	then
	    return
	fi
	echo "$_now ${0##*/}[${BASHPID}]: $* <<BB WARNING: bb_log deprecated>>" \
	    >> "$LOG_FILE" || true
    fi
}

# will be removed in the future
function log() {
    bb_log  "$@" "<<BB WARNING log is deprecated>>"
}

#@ bb_log_debug <level> <message> [..]
# logs message if $LOG_LEVEL is set and it is level <= $LOG_LEVEL
# errors: 2
function bb_log_debug() {
    (( $# > 1 )) || {
	bb_ierr "bb_log_debug called with $# instead of 2 or more args"
	return 2
    }
    local _lvl="$1"
    shift

    if [ "${LOG_LEVEL+isset}" == "isset" ]
    then
	if (( "$LOG_LEVEL" < "$1" ))
	then
	    bb_log "debug<$_lvl> $*"
	fi
    fi
}

# will be removed in the future
function log_debug() {
    bb_log_debug "$@" "<<BB WARNING: log_debug is deprecated>>"
}

#@@ flow control and error handling

#@ bb_finalize
# unregisters all traps
function bb_finalize() {
    bb_msg debug "bb_finalize called"
    _bb_traps_disable
}

#@ bb_quit [<exit-code> [<log message>] [..]]
# the safe way to clean exit a script using bash_base
function bb_quit() {
    bb_finalize
    _bb_call_exit_hooks

    # no message
    if (( $# < 2 ))
    then
	bb_msg debug "bb_quit called"
	exit "${1-0}"
    fi

    local _retval="$1"
    shift
    bb_msg info "quitting: $*"
    exit "$_retval"
}

#@ bb_exec <cmd> [arg [..]]
# execs cmd with args if specified
function bb_exec() {
    (( $# > 0 )) || {
	bb_ierr "bb_version called with $# instead of 1 or more args"
	return 2
    }

    bb_finalize
    _bb_call_exit_hooks

    bb_msg debug "bb_exec $*"

    exec "$@" || bb_fatal "bb_exec failed to exec: $*"
}

# DEPRECATED will be removed in the future
function fatal() {
    bb_fatal "$* <<BB WARNING: fatal deprecated>>"
}

function _bb_err_exit_trap() {
    local _status=$?
    local _traptype="$1"

    case "$_traptype" in
	ERR|EXIT)
	    bb_msg emerg "_bb_err_exit_trap: unexpected exit/error" \
		"near '$BASH_COMMAND', status=$_status, trap=$_traptype"
	    bb_quit "$_status"
	    ;;
	*)
	    bb_msg err "_bb_err_exit_trap: internal error," \
		"unknown trap type: $_traptype"
    esac
}

# _bb_traps_setup
# sets up traps for ERR and EXIT to handle unexpected errors and exits
function _bb_traps_setup() {
    trap '_bb_err_exit_trap "ERR"' ERR
    trap '_bb_err_exit_trap "EXIT"' EXIT
}

#@ bb_traps_setup
# will be removed in the future
function bb_traps_setup() {
    bb_msg warning \
	"<<BB WARNING: bb_traps_setup is deprecated and does nothing>>"
}

# _bb_traps_disable
# disables traps
function _bb_traps_disable() {
    trap - ERR
    trap - EXIT
}

#@ bb_traps_disable
# will be removed in the future
function bb_traps_disable() {
    bb_msg warning \
	"<<BB WARNING bb_traps_disable is deprecated and does nothing>>"
}

#@ bb_version <ret-vn>
# returns the version of bash base
# errors: 2, 3
function bb_version() {
    (( $# == 1 )) || {
	bb_ierr "bb_version called with $# instead of 1 arg"
	return 2
    }

    local _ret_vn="$1"

    bb_validate_vn bb_version "$_ret_vn" || return 3

    printf -v "$_ret_vn" "0.4.13"
}

# _bash_base_call_exit_hooks
# calls all exit hooks sorted by hooknames
function _bb_call_exit_hooks() {
    local _hookname

    # check if any hooks are registered
    (( ${#bb__exithooks[@]} > 0 )) || return 0

    for _hookname in "${!bb__exithooks[@]}"
    do
	bb_msg debug "_bb_call_exit_hooks:" \
	    "calling ${bb__exithooks[$_hookname]}"
	"${bb__exithooks[$_hookname]}" || \
	    bb_msg warn "exit hook '$_hookname' failed with exit code $?"
    done
}

#@ bb_register_exit_hook [ <hook-name> ] <function-name>
# registers a function to be called on exit, it is not possible to
# pass arguments to that function
# hooks will be called sorted by hooknames
# note: the default hook-name is the function-name
function bb_register_exit_hook() {
    (( $# == 1 || $# == 2 )) || {
	bb_ierr "bb_register_exit_hook called with $# insted of 1 or 2 args"
	return 2
    }

    local _hookname="$1"
    local _hookfun

    if (( $# == 2 ))
    then
	_hookfun="$2"
    else
	_hookfun="$_hookname"
    fi

    bb__exithooks["$_hookname"]="$_hookfun"
}

#@ bb_unregister_exit_hook <hookn-ame>
# removes a registered exit hook
# errors: 2
function bb_unregister_exit_hook() {
    (( $# == 1 )) || {
	bb_ierr "bb_unregister_exit_hook called with $# insted of 1 arg"
	return 2
    }

    local _hookname="$1"

    unset bb__exithooks["$_hookname"]
}

#@ bb_load_lib <name>
# sources file for lib specified
# if name is no absolute or local path it tries:
# /usr/libexec/<name>.bash
# /usr/libexec/<name>/<name>.bash
# errors: 2, 5, 8, 91 (lib not found)
function bb_load_lib() {
    (( $# == 1 )) || {
	bb_ierr "bb_load_lib called with $# insted of 1 arg"
	return 2
    }

    local _l="$1"

    if [[ "$_l" =~ ^\.?\.?/ ]] # is path absolute, local or relative?
    then
	if [ -r "$_l" ]
	then
	    source "$_l" || {
		bb_msg warning "bb_load_lib: error sourcing: $_l"
		return 5
	    }
	    return 0
	else
	    bb_ierr "bb_load_lib: no such absolute path: $_l"
	    return 8
	fi
    fi

    local _f
    for _f in "/usr/libexec/$_l.bash" "/usr/libexec/$_l/$_l.bash"
    do
	if [ -r "$_f" ]
	then
	    source "$_f" || {
		bb_ierr "bb_load_lib: error sourcing: $_f"
		return 5
	    }
	    return # library loaded successfully
	else
	    if [ -f "/usr/libexec/$_l.bash" ]
	    then
		bb_ierr "bb_load_lib: no access rights to: $_f"
		return 9
	    fi
	fi
    done

    bb_msg warning "bb_load_lib: no such lib found: $_l"
    return 91
}

#@ bb_load_libs <name> [..]
# load one or more libs but continue if some fail to load
# errors: 2, 91 (not all libs loaded)
function bb_load_libs() {
    (( $# > 0 )) || {
	bb_ierr "bb_msg called with $# instead of 1 or more args"
	return 2
    }

    local _retval=0

    while (( $# > 0 ))
    do
	bb_load_lib "$1" || _retval=91
	shift
    done

    return "$_retval"
}

#@ bb_require_libs <name> [..]
# load one or more libs and raise fatal error if any fail
# errors: 2, bb_fatal on any other error
function bb_require_libs() {
    (( $# > 0 )) || {
	bb_ierr "bb_msg called with $# instead of 1 or more args"
	return 2
    }

    while (( $# > 0 ))
    do
	bb_load_lib "$1" || bb_fatal "bb_require_libs: failed to load: $1"
	shift
    done
}

function _bb_main() {
    set -o errexit
    set -o pipefail
    set -o nounset
    set -o errtrace

    # default internal error mode
    declare -g bb__ierrmode="err"

    unset bb__exithooks
    declare -g -A bb__exithooks=()

    _bb_traps_setup

    # configure default message to stderr (with stack trace if >= crit)
    # shellcheck disable=SC2119
    bb_msg_clear_targets
    bb_msg_set_target default_info_err stderr info:err prog notrace
    bb_msg_set_target default_crit_emerg stderr crit:emerg prog trace
}

_bb_main
